unit gfx_colorquantizer;
{***********************************************************************
Unit gfx_colorquantizer.PAS v1.5 0801
    (c) by Andreas Moser, amoser@amoser.de

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_colorquantizer is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    algorythms are based on or taken from the sources written by
    Earl F. Glynn, efg's computer lab

********************************************************************************}


interface
uses gfx_basedef, gfx_dib,gfx_syswrapper;


// -----------------------------------------------------------------------------
//
//			ColorQuantizer
//
// -----------------------------------------------------------------------------

  type
    TOctreeNode = class;  // forward definition so TReducibleNodes can be declared

    TReducibleNodes = array[0..7] OF TOctreeNode;

    TOctreeNode     =
      class(TObject)
        IsLeaf      :  BOOLEAN;
        PixelCount  :  INTEGER;
        RedSum      :  INTEGER;
        GreenSum    :  INTEGER;
        BlueSum     :  INTEGER;
        Next        :  TOctreeNode;
        Child       :  TReducibleNodes;

        constructor Create (const Level         :  INTEGER;
                            const ColorBits     :  INTEGER;
                            var   LeafCount     :  INTEGER;
                            var   ReducibleNodes:  TReducibleNodes);
        destructor  Destroy;  override;

      end;

    TColorQuantizer =
      class(TOBject)
        private
          FTree          :  TOctreeNode;
          FLeafCount     :  INTEGER;
          FReducibleNodes:  TReducibleNodes;
          FMaxColors     :  INTEGER;
          FColorBits     :  INTEGER;

        PROTECTED
          procedure   AddColor(var   Node          :  TOctreeNode;
                               const r,g,b         :  BYTE;
                               const ColorBits     :  INTEGER;
                               const Level         :  INTEGER;
                               var   LeafCount     :  INTEGER;
                               var   ReducibleNodes:  TReducibleNodes);
          procedure   DeleteTree(var Node:  TOctreeNode);
          procedure   GetPaletteColors(const Node        :  TOctreeNode;
                                       var   RGBQuadarray:  TRGBQuadarray;
                                       var   Index       :  INTEGER);
          procedure   ReduceTree(const ColorBits:  INTEGER;
                                 var   LeafCount:  INTEGER;
                                 var   ReducibleNodes:  TReducibleNodes);

        public
          constructor Create(const MaxColors:  INTEGER; const ColorBits:  INTEGER);
          destructor  Destroy;  override;

          procedure   GetColorTable(var RGBQuadarray:  TRGBQuadarray);
          function    ProcessImage(const Handle:  TXHandle):  BOOLEAN;
          function    ProcessDIBImage(const DIB: TDIBReader): boolean;

          property    ColorCount:  INTEGER  READ FLeafCount;

      end;

implementation

//------------------------------------------------------------
// Color Quantization
//------------------------------------------------------------

// -----------------------------------------------------------------------------
//
//			Create
//
// -----------------------------------------------------------------------------
  constructor TOctreeNode.Create (const Level         :  INTEGER;
                                  const ColorBits     :  INTEGER;
                                  var   LeafCount     :  INTEGER;
                                  var   ReducibleNodes:  TReducibleNodes);
    var
      i:  INTEGER;
  begin
    PixelCount  := 0;
    RedSum      := 0;
    GreenSum    := 0;
    BlueSum     := 0;
    for i := Low(Child) to High(Child) do
      Child[i] := nil;

    IsLeaf := (Level = ColorBits);
    if   IsLeaf
    then begin
      Next := nil;
      INC(LeafCount);
    end
    else begin
      Next := ReducibleNodes[Level];
      ReducibleNodes[Level] := SELF
    end
  end {Create};


// -----------------------------------------------------------------------------
//
//			Destroy
//
// -----------------------------------------------------------------------------
  destructor TOctreeNode.Destroy;
    var
      i:  INTEGER;
  begin
      for i := Low(Child) to High(Child) do
        Child[i].Free
  end {Destroy};



// -----------------------------------------------------------------------------
//
//			Create
//
// -----------------------------------------------------------------------------
  constructor TColorQuantizer.Create(const MaxColors:  INTEGER; const ColorBits:  INTEGER);
    var
      i:  INTEGER;
  begin
    ASSERT (ColorBits <= 8);

    FTree := nil;
    FLeafCount := 0;

    // Initialize all nodes even though only ColorBits+1 of them are needed
    for i := Low(FReducibleNodes) to High(FReducibleNodes) do
      FReducibleNodes[i] := nil;

    FMaxColors := MaxColors;
    FColorBits := ColorBits
  end {Create};


// -----------------------------------------------------------------------------
//
//			Destroy
//
// -----------------------------------------------------------------------------
  destructor  TColorQuantizer.Destroy;
  begin
    if   FTree <> nil
    then DeleteTree(FTree)
  end {Destroy};


// -----------------------------------------------------------------------------
//
//			GetColorTable
//
// -----------------------------------------------------------------------------
procedure TColorQuantizer.GetColorTable(var RGBQuadarray:  TRGBQuadarray);
var
      Index:  INTEGER;
begin
    Index := 0;
    GetPaletteColors(FTree, RGBQuadarray, Index)
end {GetColorTable};


// -----------------------------------------------------------------------------
//
//			ProcessImage
//
// -----------------------------------------------------------------------------
  // Handles passed to ProcessImage should refer to DIB sections, not DDBs.
  // In certain cases, specifically when it's called upon to process 1, 4, or
  // 8-bit per pixel images on systems with palettized display adapters,
  // ProcessImage can produce incorrect results if it's passed a handle to a
  // DDB.
function TColorQuantizer.ProcessImage(const Handle:  TXHandle):  BOOLEAN;
const
      MaxPixelCount = 1048576;  // 2^20 shouldn't be much of a limit here

var
      Bytes     :  INTEGER;
      DIBSection:  TXDIBSection;

    // Process 1, 4, or 8-bit DIB:
    // The strategy here is to use GetDIBits to convert the image into
    // a 24-bit DIB one scan line at a time.  A pleasant side effect
    // of using GetDIBits in this manner is that RLE-encoded 4-bit and
    // 8-bit DIBs will be uncompressed.

    // Implemented as in article, but doesn't work (yet) as I would expect.
    procedure ProcessLowBitDIB;
      var
        BitmapInfo   :  TxBitmapInfo;
        DeviceContext:  hxDC;
        i            :  INTEGER;
        j            :  INTEGER;
        ScanLine     :  pRGBarray;
    begin
      GetMem(ScanLine, 3*DIBSection.dsBmih.biWidth);
      try
        xZeroMemory(@BitmapInfo, SizeOf(BitmapInfo));
        with BitmapInfo do
        begin
          bmiHeader.biSize        := SizeOf(TxBitmapInfo);
          bmiHeader.biWidth       := DIBSection.dsBmih.biWidth;
          bmiHeader.biHeight      := DIBSection.dsBmih.biHeight;
          bmiHeader.biPlanes      := 1;
          bmiHeader.biBitCount    := 24;
          bmiHeader.biCompression := xBI_RGB;

        end;

        DeviceContext := xGetDC(0);
        try
          for j := 0 to DIBSection.dsBmih.biHeight-1 do
          begin

            xGetDIBits (DeviceContext, Handle, j, 1, ScanLine, BitmapInfo, xDIB_RGB_COLORS);
            for i := 0 to DIBSection.dsBmih.biWidth-1 do
            begin
              with Scanline[i] do
                AddColor(FTree, rgbtRed, rgbtGreen, rgbtBlue,
                         FColorBits, 0, FLeafCount, FReducibleNodes);

              while FLeafCount > FMaxColors do
                ReduceTree(FColorbits, FLeafCount, FReducibleNodes)

            end
          end

        finally
          xReleaseDC(0, DeviceContext);
        end

      finally
        FreeMem(ScanLine)
      end
    end {ProcessLowBitDIB};


    procedure Process16BitDIB;
    begin
      // Not yet implemented
    end {Process16BitDIB};


    procedure Process24BitDIB;
      var
        i       :  INTEGER;
        j       :  INTEGER;
        ScanLine:  pRGBarray;
    begin
      Scanline := pRGBarray(DIBSection.dsBm.bmBits);
      for j := 0 to DIBSection.dsBmih.biHeight-1 do
      begin

        for i := 0 to DIBSection.dsBmih.biWidth-1 do
        begin
          with Scanline[i] DO
            AddColor(FTree, rgbtRed, rgbtGreen, rgbtBlue,
                     FColorBits, 0, FLeafCount, FReducibleNodes);

          WHILE FLeafCount > FMaxColors DO
            ReduceTree(FColorbits, FLeafCount, FReducibleNodes)

        end;

        ScanLine := pRGBarray(INTEGER(Scanline) + DIBSection.dsBm.bmWidthBytes);
      end
    end {Process24BitDIB};


    procedure Process32BitDIB;
    begin
      // Not yet implemented
    end {Process32BitDIB};

begin
    RESULT := FALSE;

    Bytes := xGetObject(Handle, SizeOF(DIBSection), @DIBSection);

    IF   Bytes > 0   // Invalid Bitmap if Bytes = 0
    THEN begin
//    PadBytes := DIBSECTION.dsBm.bmWidthBytes -
//                (((DIBSection.dsBmih.biWidth * DIBSection.dsBmih.biBitCount) + 7) DIV 8);
      ASSERT (DIBSection.dsBmih.biHeight < MaxPixelCount);
      ASSERT (DIBSection.dsBmih.biWidth  < MaxPixelCount);

      CASE  DIBSection.dsBmih.biBitCount OF
         1:  ProcessLowBitDIB;
         4:  ProcessLowBitDIB;
         8:  ProcessLowBitDIB;
        16:  Process16bitDIB;
        24:  Process24bitDIB;
        32:  Process32bitDIB
        else
          // Do nothing.  Default RESULT is already FALSE
      end

    end
end ;



function TColorQuantizer.ProcessDIBImage(const DIB: TDIBReader): boolean;
var
  i			,
  j			: integer;
  ScanLine		: pointer;
  Pixel			: PRGBTriple;
begin
  Result := True;

  for j := 0 to DIB.Bitmap.Height-1 do
  begin
    Scanline := DIB.Scanline[j];
    Pixel := ScanLine;
    for i := 0 to DIB.Bitmap.Width-1 do
    begin
      with Pixel^ do
        AddColor(FTree, rgbtRed, rgbtGreen, rgbtBlue,
                 FColorBits, 0, FLeafCount, FReducibleNodes);

      while FLeafCount > FMaxColors do
        ReduceTree(FColorbits, FLeafCount, FReducibleNodes);
      inc(Pixel);
    end;
  end;
end;

// -----------------------------------------------------------------------------
//
//			AddColor
//
// -----------------------------------------------------------------------------
procedure TColorQuantizer.AddColor(var   Node          :  TOctreeNode;
                                     const r,g,b         :  BYTE;
                                     const ColorBits     :  INTEGER;
                                     const Level         :  INTEGER;
                                     var   LeafCount     :  INTEGER;
                                     var   ReducibleNodes:  TReducibleNodes);
const
      Mask:  array[0..7] OF BYTE = ($80, $40, $20, $10, $08, $04, $02, $01);

var
      Index    :  INTEGER;
      Shift    :  INTEGER;
begin
    // If the node doesn't exist, create it.
    IF   Node = NIL
    THEN Node := TOctreeNode.Create(Level, ColorBits, LeafCount, ReducibleNodes);

    IF   Node.IsLeaf
    THEN begin
      INC(Node.PixelCount);
      INC(Node.RedSum,   r);
      INC(Node.GreenSum, g);
      INC(Node.BlueSum,  b)
    end
    else begin
      // Recurse a level deeper if the node is not a leaf.
      Shift := 7 - Level;

      Index := (((r AND mask[Level]) SHR Shift) SHL 2)  OR
               (((g AND mask[Level]) SHR Shift) SHL 1)  OR
                ((b AND mask[Level]) SHR Shift);
      AddColor(Node.Child[Index], r, g, b, ColorBits, Level+1,
               LeafCount, ReducibleNodes)
    end

end ;



// -----------------------------------------------------------------------------
//
//			DeleteTree
//
// -----------------------------------------------------------------------------
procedure TColorQuantizer.DeleteTree(var Node:  TOctreeNode);
var
      i        :  INTEGER;
begin
    for i := Low(TReducibleNodes) TO High(TReducibleNodes) DO
    begin
      IF   Node.Child[i] <> NIL
      THEN DeleteTree(Node.Child[i]);
    end;

    Node.Free;

    Node := NIL;
end;


// -----------------------------------------------------------------------------
//
//			GetPaletteColors
//
// -----------------------------------------------------------------------------
procedure TColorQuantizer.GetPaletteColors(const Node        :  TOctreeNode;
                                             var   RGBQuadarray:  TRGBQuadarray;
                                             var   Index       :  INTEGER);
var
      i:  INTEGER;
begin
    if Assigned(Node) then IF   Node.IsLeaf
    THEN begin
      with RGBQuadarray[Index] DO
      begin
        TRY
          rgbRed   := BYTE(Node.RedSum   DIV Node.PixelCount);
          rgbGreen := BYTE(Node.GreenSum DIV Node.PixelCount);
          rgbBlue  := BYTE(Node.BlueSum  DIV Node.PixelCount)
        EXCEPT
          rgbRed   := 0;
          rgbGreen := 0;
          rgbBlue  := 0
        end;

        rgbReserved := 0
      end;
      INC(Index)
    end
    else begin
      for i := Low(Node.Child) TO High(Node.Child) DO
      begin
        IF   Node.Child[i] <> NIL
        THEN GetPaletteColors(Node.Child[i], RGBQuadarray, Index)
      end
    end
end;


// -----------------------------------------------------------------------------
//
//			ReduceTree
//
// -----------------------------------------------------------------------------
procedure TColorQuantizer.ReduceTree(const ColorBits:  INTEGER;
                                       var   LeafCount:  INTEGER;
                                       var   ReducibleNodes:  TReducibleNodes);
var
      BlueSum :  INTEGER;
      Children:  INTEGER;
      GreenSum:  INTEGER;
      i       :  INTEGER;
      Node    :  TOctreeNode;
      RedSum  :  INTEGER;
begin
    // Find the deepest level containing at least one reducible node
    i := Colorbits - 1;
    WHILE (i > 0) AND (ReducibleNodes[i] = NIL) DO
      DEC(i);

    // Reduce the node most recently added to the list at level i.
    Node := ReducibleNodes[i];
    ReducibleNodes[i] := Node.Next;

    RedSum   := 0;
    GreenSum := 0;
    BlueSum  := 0;
    Children := 0;

    for i := Low(ReducibleNodes) TO High(ReducibleNodes) DO
    begin
      IF   Node.Child[i] <> NIL
      THEN begin
        INC(RedSum,          Node.Child[i].RedSum);
        INC(GreenSum,        Node.Child[i].GreenSum);
        INC(BlueSum,         Node.Child[i].BlueSum);
        INC(Node.PixelCount, Node.Child[i].PixelCount);
        Node.Child[i].Free;
        Node.Child[i] := NIL;
        INC(Children)
      end
    end;

    Node.IsLeaf   := TRUE;
    Node.RedSum   := RedSum;
    Node.GreenSum := GreenSum;
    Node.BlueSum  := BlueSum;
    DEC(LeafCount, Children-1)
end;


end.
